/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/

#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/sockios.h>
#include <netinet/in.h>
#include <net/if.h>
#include <signal.h>
#include <fcntl.h>
#include <varargs.h>
#include <ctype.h>

#include <socketInterface.h>
#include <Timer.h>
#include <myConn.h>
#include "globals.h"

unsigned long  subSystem_trace_mask = 0;
int trace_level= 0; /* default: zero, i.e., no trace output            */

static Connections serverConnections;

static void sig(int signo)
{
  if ((signo == SIGTERM) || (signo == SIGINT) ||
      (signo == SIGSEGV)) {

    printf ("\n");
      
    switch (signo) {
    case SIGSEGV: {
      printf ("Segmentation fault, aborting.\n");
      exit(1);
    }
    case SIGTERM: 
      // Put all you "clean_up" code here
      break;
    case SIGINT: {
      // Ignore
      printf("Received SIGINT\n");
      serverConnections.CleanupConnections();
      exit(0);
    }
    case SIGPIPE: {
      // Ignore
      printf("Received SIGPIPE - ignoring\n");
    }
    }
  }
}


void init_signals()
{
  signal (SIGTERM, sig);
  signal (SIGINT, sig);
  signal (SIGSEGV, sig); 
}	

void printUsage() {
  cout << "Usage:: server -p port [-s <source IP>]" << endl;
}

int main (int argc, char* argv[])
{
  int srcIp = INADDR_ANY;
  int server_port = -1;
  
  char ch;
  char *cp;

  while ((ch = getopt(argc, argv, "T:p:s")) != EOF) {
    switch(ch) {
    case 'p':
      server_port = atoi(optarg);
      break;
    case 's':
      break;
     case 'T': /* Tracing level */
       for (cp=optarg; *cp; cp++) /* test for numeric option */
         if (!isdigit((int)*cp)){
	   printf("Error: trace-level must be an unsigned integer\n");
	   printUsage();
	   exit(1);
	 }
       trace_level = atoi(optarg);
    }
  }
  if (server_port == -1) {
    printUsage();
    exit(1);
  }
  cout << "Server::Port = " << server_port << endl;

  init_signals();
  chdir("/");
  
  int lFd;
  if ((lFd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    exit(0);
  }

  struct sockaddr_in serv_addr;
  bzero((char *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(srcIp);
  serv_addr.sin_port = htons(server_port);

  if (bind(lFd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    perror("Server::Can't bind to local ip address");
    exit (1);
  }

  listen(lFd, 5);

  fd_set  readfds;
  fd_set  writefds;

  /* Timer stuff */
  Timer myTimer;
  myTimer.addTimer("timer1", 15000, myConn::timer_routine);  // 15 secs
  myTimer.addTimer("timer2", 60000, myConn::timer_routine);  // 60 secs
  //myTimer.addTimer("timer3", 30000, myConn::timer_routine);  // 30 secs
  myTimer.printTimerList("Main");

  while (1) {
    struct timeval t;
    if (!myTimer.getNewTimeToWait(t)) {
      t.tv_sec = 1;  // 1 sec timeout - default
      t.tv_usec = 0;
    }
    int maxfdpl = 0;

    FD_ZERO(&readfds);
    FD_ZERO(&writefds);

    FD_SET(lFd, &readfds);
    maxfdpl = (maxfdpl < lFd) ? lFd : maxfdpl;
    
    // Get Max connection fd & set the readfd set
    serverConnections.deleteDeadConnections();

    int tempMax = serverConnections.setReadWriteFdSets(readfds, writefds);
    maxfdpl = (maxfdpl > tempMax) ? maxfdpl : tempMax;

    maxfdpl++;

    if (select(maxfdpl, (fd_set *)&readfds, (fd_set *)&writefds, (fd_set *)0, &t) < 0) {
      cout << "Error in select call" << endl;
      continue;  // we don't exit as we would like the signal handler to handle all exits
      //exit(0);
    }

    myTimer.timerFired();

    if (FD_ISSET(lFd, &readfds)) {
      struct sockaddr peer;
      int newFd = SocketInterface::acceptAndGetNewSocket(peer, lFd);
      if (!myConn::staticConn) {
	// Create "My" connection
	if (!myConn::staticConn) {
	  myConn::staticConn = new myConn(&serverConnections, SERVER, newFd);  // Server side socket
	}
	else {
	  myConn::staticConn->init(SERVER, newFd);  // Server side socket
	}
      }
    }
    serverConnections.findConnectionAndDispatch(readfds, writefds);
    // cout << "In select loop" << endl;
  }
  return 0;
}


void printDebug(char *buff, int buffLen)
{
}
