/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/

/*
  _############################################################################
  _## 
  _##  SocketInterface.h
  _## 
  _##
  _##  
  _##########################################################################
*/

#ifndef _SOCKETINTERFACE_H_
#define _SOCKETINTERFACE_H_

#include <sys/socket.h>
#include <linux/sockios.h>
#include <errno.h>
#include <tlv.h>
#include <unistd.h>
#include <fcntl.h>
#include <semaphore.h>
#include <pthread.h> 
#include <netdb.h>

extern int traceLevel;

#define MUTEX_GOOD 1
#define MUTEX_BAD  2

class SockMutex {
 public:
  // if called with 
  //   1 - means dont wait first time "wait" is called
  //   0 - means wait first time "wait" is called (default)
  SockMutex(int wait = 0) {
    status = MUTEX_GOOD;
    if (init(wait)) status = MUTEX_BAD;
  };
  ~SockMutex() {if (status == MUTEX_GOOD) sem_destroy(&mutex);};

 private:
  sem_t mutex;   // to be used if thread needs to be blocked for the reply
  int   status;
  
 public:
  void wait() {sem_wait(&mutex);};
  int  timedWait(int timeout);
  void release() {sem_post(&mutex);};
  int  increment() {if(!sem_post(&mutex)) return getVal(); return -1;};
  int  decrement() {if(!sem_trywait(&mutex)) return getVal(); return -1;}
  int  getVal() {int val = -1; if (sem_getvalue(&mutex, &val) == 0) return val; return -1;};
  int  init(int val) {
    int  pshared = 0; /* local to the current process */
    if (0 != sem_init(&mutex, pshared, val)) return 0;
    return 1;
  };
};

class transaction {
 public:
  transaction(int t_out);
  ~transaction();
  static transaction *findtid(int t);
  static transaction *getTransaction(int t_out = 0) {transaction *p = new transaction(t_out); return p;};

 private:
  SockMutex   first;
  SockMutex   second;
  static transaction *T_head;
  static transaction *T_tail;
  static int  globalTid;

  transaction *next;
  transaction *prev;

 public:
  int tid;
  int timeOut;
  int lastTypeReceived;

 public:
  int getTid() {return tid;};
  int timedWait();
  void wait(){first.wait();};
  void wakeup(){first.release();};
  void postWait(){second.wait();};
  void postWakeup(){second.release();};
  void print() {cout << "TID = " << tid << "; Sema Val = " << first.getVal() << endl;};
};

class sendBuffer {
 public:
  sendBuffer() : next(NULL), prev(NULL), buffer(NULL), bufferLen(0) {};
  sendBuffer(char *buff, int buffLen) : next(NULL), prev(NULL), buffer(NULL), bufferLen(0) {
    if (!buff) return;
    buffer = (char *)malloc(buffLen);
    memcpy(buffer, buff, buffLen);
    bufferLen = buffLen;
  };
  ~sendBuffer() {if (buffer) free(buffer);};
  void freeRecursive() {
    if (next)
      next->freeRecursive();
    else
      delete this;
  };

 public:
  sendBuffer *next;
  sendBuffer *prev;
  char *buffer;
  int bufferLen;
};

#define MAX_BUFFER_LIMIT 2000

class sendBufferList {
 public:
  sendBufferList() : bufferHead(NULL), bufferTail(NULL), numBuffers(0) {};
  ~sendBufferList();  // Removes the entire list

 private:
  sendBuffer *bufferHead;
  sendBuffer *bufferTail;

 public:
  int numBuffers;

 public:
  int add(char *buff, int buffLen);
  sendBuffer *remove(sendBuffer *sb); // returns the next one
  int  flushToSocket(int fd);
  // This is to get rid of any/all send buffers that may be queued up
  void cleanup();
};

#define CONNECTED 1
#define CONNECT_PENDING 2
#define DISCONNECTED 3
#define DEAD 4
#define SOCKET_FULL 5
  
#define CLIENT 1
#define SERVER 2

class Connection;

class SocketInterface {
 public:  //Constructors
  SocketInterface(int clientOrServer) {
    int sock = createNonBlockingSocket();
    init(clientOrServer, sock);
    readBuffer = processedDataPtr = NULL;
    processedDataLen = readBufferLen = 0;
    State = DISCONNECTED; connecting = false; lastConnectTime = 0;
  };
  SocketInterface(int clientOrServer, int sock, int state = DISCONNECTED) {
    init(clientOrServer, sock);
    readBuffer = processedDataPtr = NULL;
    processedDataLen = readBufferLen = 0;
    State = state; connecting = false; lastConnectTime = 0;
  };
  virtual ~SocketInterface() {if (readBuffer) free (readBuffer); if (fd != -1) ::close(fd);};

 public:
  void init(int clientOrServer, int sock) {fd = sock; clientServer = clientOrServer;};
  int  getFd() {return fd;};
  void setFd(int sock) {fd = sock;};
  bool flush();
  bool buffersToSend();
  bool getTlv();
  bool read2K();
  int  shrinkReadBuffer();
  bool bufferExceeded();
  bool sendTlv(tlv &inTlv);
  bool sendTlv(int type, int length, char *message);
  bool isClient() {if(clientServer == CLIENT) return true; return false;};
  bool isServer() {if(clientServer == SERVER) return true; return false;};
  bool isConnected() {if (State == CONNECTED || State == SOCKET_FULL) return true; return false;};
  bool isSocketFull() {if (State == SOCKET_FULL) return true; return false;};
  bool dead() {if (State == DEAD) return true; return false;};
  static int verifyIpAddress(char *str);
  static int createNonBlockingSocket();
  static int acceptAndGetNewSocket(struct sockaddr &peer, int listenFd);
  static int connectToServer(int port, Connection *conn, char *ipAddress, char *errorString);
  void connectPending() {State = CONNECT_PENDING;};
  bool isConnectPending() {if (State == CONNECT_PENDING) return true; return false;};
  void connected() {State = CONNECTED; connecting = false;};
  void stateString(char *inputbuffer) {
    const char *p = ((State == CONNECTED) ? "CONNECTED" : 
		     ((State == CONNECT_PENDING) ? "CONNECT_PENDING" : 
		      ((State == DISCONNECTED) ? "DISCONNECTED" : 
		       ((State == SOCKET_FULL) ? "SOCKET_FULL" : 
			((State == DEAD) ? "DEAD" : "UNKNOWN")))));
    strcpy(inputbuffer, p);
  };
  /*
  const char *stateString(void) {
    return ((State == CONNECTED) ? "CONNECTED" : 
	    ((State == CONNECT_PENDING) ? "CONNECT_PENDING" : 
	     ((State == DISCONNECTED) ? "DISCONNECTED" : 
	      ((State == SOCKET_FULL) ? "SOCKET_FULL" : 
	       ((State == DEAD) ? "DEAD" : "UNKNOWN")))));
  };
  */  
  void printState(void) {char buffer[80]; stateString(buffer); cout << buffer;};

  
 protected:
  void socketCleanup();
  bool Send(char *buff, int buffLen);  // This can only be called by the Connection layer
  virtual void disconnected() = 0;     // Pure virtual
  time_t lastConnectTime;

 private:
  char *readBuffer;
  int   readBufferLen;
  char *processedDataPtr;
  int  processedDataLen;
  bool connecting;

  sendBufferList m_sendBuffers;

 public:
  int   State;
  
 protected:
  int  fd;
  tlv  receiveTlv;
  int  clientServer;
};

class Connections {
 public:
  Connections(int dbg = 0, char *name = NULL);
  virtual ~Connections() {CleanupConnections();};
  Connection *connectionsHead;  
  Connection *connectionsTail;  
  int        numConnections;

 private:
  SockMutex   lock;
  int         count;
  char        listName[80];
  int         connectionsDebug;

 public:
  void printMutex(char *pre, char *pre2);
  void grabMutexLock(char *pre);
  void releaseMutexLock(char *pre);
  void add(Connection *conn);
  void remove(Connection *conn);
  void CleanupConnections();
  void removeAllChildren();   // Used when temporarily kept connections need to be removed from the list

  int  getNumConnections() {return numConnections;};
  virtual void deleteDeadConnections();
  int  setReadWriteFdSets(fd_set  &readfds, fd_set  &writefds);
  int  setFdSet(fd_set &fds);
  void findConnectionAndDispatch(fd_set  &readfds, fd_set  &writefds);
  //  Connection *getUnusedConnection();
};

class Connection : public SocketInterface  {
 public:
  Connection(Connections *list, int clientOrServer, char *name = NULL);
  Connection(Connections *list, int clientOrServer, int fd, char *name = NULL);
  virtual ~Connection();

 private:
  Connections *connectionsList;
  SockMutex   lock;
  char        connName[80];

 public:
  Connection *next;
  Connection *prev;

 public:
  void         setName(char *name) {
    if (name) if (strlen(name) < 80) strcpy(connName, name); else strncpy(connName, name, 79);
  };
  char        *getName() {return connName;};
  int          setFdSet(fd_set &fds);
  int          Connect(struct sockaddr *serv_addr);
  void         disconnect() {disconnected();};
  virtual void userConnectionEstablished() = 0;
  virtual void processTLV() = 0;  // pure virtual
  virtual void userDisconnectProcessing() = 0;  // pure virtual
  virtual void userOkToSend() = 0; // pure virtual
  virtual void disconnected();
  // Used before deleting the connection object
  int     checkRefCount() {return lock.getVal();};
  // Use these functions to access the conn object from a different thread
  int     incrementRefCount() {return lock.increment();};
  int     decrementRefCount() {return lock.decrement();};
};
 
#endif
